"""@package ecohydrolib.dbf
    
@brief Methods for reading and write DBF data files
@brief Adapted from: http://code.activestate.com/recipes/362715-dbf-reader-and-writer/

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Raymond Hettinger
@author Brian Miles <brian_miles@unc.edu>
"""
import struct, datetime, decimal, itertools

def dbfreader(f):
    """ Returns an iterator over records in a Xbase DBF file.

        @param f A file descriptor
        @return An iterator of records in the file
    
        The first row returned contains the field names.
        The second row contains field specs: (type, size, decimal places).
        Subsequent rows contain the data records.
        If a record is marked as deleted, it is skipped.
    
        File should be opened for binary reads.
    """
    # See DBF format spec at:
    #     http://www.pgts.com.au/download/public/xbase.htm#DBF_STRUCT

    numrec, lenheader = struct.unpack('<xxxxLH22x', f.read(32))    
    numfields = (lenheader - 33) // 32

    fields = []
    for fieldno in xrange(numfields):
        name, typ, size, deci = struct.unpack('<11sc4xBB14x', f.read(32))
        name = name.replace('\0', '')       # eliminate NULs from string   
        fields.append((name, typ, size, deci))
    yield [field[0] for field in fields]
    yield [tuple(field[1:]) for field in fields]

    terminator = f.read(1)
    assert terminator == '\r'

    fields.insert(0, ('DeletionFlag', 'C', 1, 0))
    fmt = ''.join(['%ds' % fieldinfo[2] for fieldinfo in fields])
    #fmt = ''.join(['%ds' % (fieldinfo[2] + (fieldinfo[3] * 256)) for fieldinfo in fields]) # Support fields with lengths greater than 255?
    fmtsiz = struct.calcsize(fmt)
    for i in xrange(numrec):
        record = struct.unpack(fmt, f.read(fmtsiz))
        if record[0] != ' ':
            continue                        # deleted record
        result = []
        for (name, typ, size, deci), value in itertools.izip(fields, record):
            if name == 'DeletionFlag':
                continue
            if typ == "N":
                value = value.replace('\0', '').lstrip()
                if value == '':
                    value = 0
                elif deci:
                    value = decimal.Decimal(value)
                else:
                    value = int(value)
            elif typ == 'D':
                value = value.lstrip() # Catch empty fields
                if value != '':
                    y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
                    value = datetime.date(y, m, d)
                else:
                    value = datetime.date(1900, 1, 1) # An empty date
            elif typ == 'L':
                value = (value in 'YyTt' and 'T') or (value in 'NnFf' and 'F') or '?'
            elif typ == 'F':
                value = float(value)
            elif typ == 'C':
                #value = value.lstrip()
                value = value.strip()
            result.append(value)
        yield result


def dbfwriter(f, fieldnames, fieldspecs, records):
    """ Return a string suitable for writing directly to a binary dbf file.

        @param f File descriptor, should be open for writing in a binary mode.
        @param fieldnames List of field names, should be no longer than ten characters and not include \x00.
        @param fieldspecs List of field specifications, in the form (type, size, deci) where
            type is one of:
                C for ascii character data
                M for ascii character memo data (real memo fields not supported)
                D for datetime objects
                N for ints or decimal objects
                L for logical values 'T', 'F', or '?'
            size is the field width
            deci is the number of decimal places in the provided decimal object
        @param records An iterable over the records (sequences of field values).
    """
    # header info
    ver = 3
    now = datetime.datetime.now()
    yr, mon, day = now.year-1900, now.month, now.day
    numrec = len(records)
    numfields = len(fieldspecs)
    lenheader = numfields * 32 + 33
    lenrecord = sum(field[1] for field in fieldspecs) + 1
    hdr = struct.pack('<BBBBLHH20x', ver, yr, mon, day, numrec, lenheader, lenrecord)
    f.write(hdr)
                      
    # field specs
    for name, (typ, size, deci) in itertools.izip(fieldnames, fieldspecs):
        name = name.ljust(11, '\x00')
        fld = struct.pack('<11sc4xBB14x', name, typ, size, deci)
        f.write(fld)

    # terminator
    f.write('\r')

    # records
    for record in records:
        f.write(' ')                        # deletion flag
        for (typ, size, deci), value in itertools.izip(fieldspecs, record):
            if typ == "N":
                value = str(value).rjust(size, ' ')
            elif typ == 'D':
                value = value.strftime('%Y%m%d')
            elif typ == 'L':
                value = str(value)[0].upper()
            else:
                value = str(value)[:size].ljust(size, ' ')
            assert len(value) == size
            f.write(value)

    # End of file
    f.write('\x1A')


# -------------------------------------------------------
# Example calls
if __name__ == '__main__':
    import sys, csv
    from cStringIO import StringIO
    from operator import itemgetter

    # Read a database
    filename = '/pydev/databases/orders.dbf'      
    if len(sys.argv) == 2:
        filename = sys.argv[1]
    f = open(filename, 'rb')
    db = list(dbfreader(f))
    f.close()
    for record in db:
        print record
    fieldnames, fieldspecs, records = db[0], db[1], db[2:]

    # Alter the database
    del records[4]
    records.sort(key=itemgetter(4))

    # Remove a field
    del fieldnames[0]
    del fieldspecs[0]
    records = [rec[1:] for rec in records]

    # Create a new DBF
    f = StringIO()
    dbfwriter(f, fieldnames, fieldspecs, records)

    # Read the data back from the new DBF
    print '-' * 20    
    f.seek(0)
    for line in dbfreader(f):
        print line
    f.close()

    # Convert to CSV
    print '.' * 20    
    f = StringIO()
    csv.writer(f).writerow(fieldnames)    
    csv.writer(f).writerows(records)
    print f.getvalue()
    f.close()

# Example Output
"""
['ORDER_ID', 'CUSTMR_ID', 'EMPLOY_ID', 'ORDER_DATE', 'ORDER_AMT']
[('C', 10, 0), ('C', 11, 0), ('C', 11, 0), ('D', 8, 0), ('N', 12, 2)]
['10005     ', 'WALNG      ', '555        ', datetime.date(1995, 5, 22), Decimal("173.40")]
['10004     ', 'BMARK      ', '777        ', datetime.date(1995, 5, 18), Decimal("3194.20")]
['10029     ', 'SAWYH      ', '777        ', datetime.date(1995, 6, 29), Decimal("97.30")]
['10013     ', 'RITEB      ', '777        ', datetime.date(1995, 6, 2), Decimal("560.40")]
['10024     ', 'RATTC      ', '444        ', datetime.date(1995, 6, 21), Decimal("2223.50")]
['10018     ', 'RATTC      ', '444        ', datetime.date(1995, 6, 12), Decimal("1076.05")]
['10025     ', 'RATTC      ', '444        ', datetime.date(1995, 6, 23), Decimal("185.80")]
['10038     ', 'OLDWO      ', '111        ', datetime.date(1995, 7, 14), Decimal("863.96")]
['10002     ', 'MTIME      ', '333        ', datetime.date(1995, 5, 16), Decimal("731.80")]
['10007     ', 'MORNS      ', '444        ', datetime.date(1995, 5, 24), Decimal("1405.00")]
['10026     ', 'MORNS      ', '555        ', datetime.date(1995, 6, 26), Decimal("17.40")]
['10030     ', 'LILLO      ', '111        ', datetime.date(1995, 7, 3), Decimal("909.91")]
['10022     ', 'LAPLA      ', '111        ', datetime.date(1995, 6, 19), Decimal("671.50")]
['10035     ', 'HIGHG      ', '111        ', datetime.date(1995, 7, 11), Decimal("1984.83")]
['10033     ', 'FOODG      ', '333        ', datetime.date(1995, 7, 6), Decimal("3401.32")]
--------------------
['CUSTMR_ID', 'EMPLOY_ID', 'ORDER_DATE', 'ORDER_AMT']
[('C', 11, 0), ('C', 11, 0), ('D', 8, 0), ('N', 12, 2)]
['MORNS      ', '555        ', datetime.date(1995, 6, 26), Decimal("17.40")]
['SAWYH      ', '777        ', datetime.date(1995, 6, 29), Decimal("97.30")]
['WALNG      ', '555        ', datetime.date(1995, 5, 22), Decimal("173.40")]
['RATTC      ', '444        ', datetime.date(1995, 6, 23), Decimal("185.80")]
['RITEB      ', '777        ', datetime.date(1995, 6, 2), Decimal("560.40")]
['LAPLA      ', '111        ', datetime.date(1995, 6, 19), Decimal("671.50")]
['MTIME      ', '333        ', datetime.date(1995, 5, 16), Decimal("731.80")]
['OLDWO      ', '111        ', datetime.date(1995, 7, 14), Decimal("863.96")]
['LILLO      ', '111        ', datetime.date(1995, 7, 3), Decimal("909.91")]
['RATTC      ', '444        ', datetime.date(1995, 6, 12), Decimal("1076.05")]
['MORNS      ', '444        ', datetime.date(1995, 5, 24), Decimal("1405.00")]
['HIGHG      ', '111        ', datetime.date(1995, 7, 11), Decimal("1984.83")]
['BMARK      ', '777        ', datetime.date(1995, 5, 18), Decimal("3194.20")]
['FOODG      ', '333        ', datetime.date(1995, 7, 6), Decimal("3401.32")]
....................
CUSTMR_ID,EMPLOY_ID,ORDER_DATE,ORDER_AMT
MORNS      ,555        ,1995-06-26,17.40
SAWYH      ,777        ,1995-06-29,97.30
WALNG      ,555        ,1995-05-22,173.40
RATTC      ,444        ,1995-06-23,185.80
RITEB      ,777        ,1995-06-02,560.40
LAPLA      ,111        ,1995-06-19,671.50
MTIME      ,333        ,1995-05-16,731.80
OLDWO      ,111        ,1995-07-14,863.96
LILLO      ,111        ,1995-07-03,909.91
RATTC      ,444        ,1995-06-12,1076.05
MORNS      ,444        ,1995-05-24,1405.00
HIGHG      ,111        ,1995-07-11,1984.83
BMARK      ,777        ,1995-05-18,3194.20
FOODG      ,333        ,1995-07-06,3401.32
"""




#!/usr/bin/env python
"""@package GetNLCDForDEMExtent

@brief Extract a tile, the extent derived from the DEM of the project, of NLCD 2006 data 
from a locally stored copy of the entire NLCD 2006 dataset. 

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  

Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GDAL/OGR', 'PATH_OF_GDAL_WARP'
   'NLCD', 'PATH_OF_NLCD2006'

2. following metadata entry(ies) must be present in the manifest section of the metadata associated with the project directory:
   dem

3. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   bbox_wgs84
   dem_res_x
   dem_res_y
   dem_srs

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   landcover [the name of the landcover raster]
   
2. Will write the following entry(ies) to the study area section of metadata associated with the project directory:
   landcover_type=NLCD2006

Usage:
@code
python ./GetNLCDForDEMExtent.py -p /path/to/project_dir
@endcode
"""
import os
import sys
import errno
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import extractTileFromRaster
from ecohydrolib.spatialdata.utils import extractTileFromRasterByRasterExtent
from ecohydrolib.spatialdata.utils import resampleRaster
from ecohydrolib.spatialdata.utils import deleteGeoTiff
from ecohydrolib.spatialdata.utils import getRasterExtentAsBbox
from ecohydrolib.nlcd.daacquery import getNLCDForBoundingBox
from ecohydrolib.nlcd.daacquery import HOST

# Handle command line options
parser = argparse.ArgumentParser(description='Get NLCD data (in GeoTIFF format) for DEM extent from a local copy of the entire NLCD 2006 dataset.')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-s', '--source', dest='source', required=False, choices=['local', 'wcs'], default='wcs',
                    help='Source for NLCD data')
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the NLCD file to be written.  File extension ".tif" will be added.')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GDAL/OGR', 'PATH_OF_GDAL_WARP'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_GDAL_WARP'))

if args.outfile:
    outfile = args.outfile
else:
    outfile = "NLCD"
tileFilename = "%s.tif" % (outfile)

# Get name of DEM raster
manifest = GenericMetadata.readManifestEntries(context)
demFilename = manifest['dem']
demFilepath = os.path.join(context.projectDir, demFilename)
demFilepath = os.path.abspath(demFilepath)
bbox = getRasterExtentAsBbox(demFilepath)

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
outputrasterresolutionX = studyArea['dem_res_x']
outputrasterresolutionY = studyArea['dem_res_y']
srs = studyArea['dem_srs']

if args.source == 'local':

    nlcdURL = 'http://gisdata.usgs.gov/TDDS/DownloadFile.php?TYPE=nlcd2006&FNAME=NLCD2006_landcover_4-20-11_se5.zip'
    nlcdRaster = context.config.get('NLCD', 'PATH_OF_NLCD2006')
    if not os.access(nlcdRaster, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read NLCD raster %s" % (nlcdRaster,))
    nlcdRaster = os.path.abspath(nlcdRaster)
    
    sys.stdout.write('Extracting tile from local NLCD data...')
    sys.stdout.flush()
    extractTileFromRasterByRasterExtent(context.config, context.projectDir, demFilepath, nlcdRaster, tileFilename)
    sys.stdout.write('done\n')

else:
    # Download NLCD from WCS
    sys.stdout.write("Downloading NLCD via WCS from %s..." % (HOST,) )
    sys.stdout.flush()
    (returnCode, nlcdURL) = getNLCDForBoundingBox(context.config, context.projectDir, tileFilename, bbox=bbox, 
                                                  resx=outputrasterresolutionX, resy=outputrasterresolutionY, 
                                                  coverage='NLCD2006', srs=srs)
    assert(returnCode)
    sys.stdout.write('done\n')

# Write metadata
GenericMetadata.writeStudyAreaEntry(context, "landcover_type", "NLCD2006")

# Write provenance
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'landcover'
asset.dcIdentifier = tileFilename
asset.dcSource = nlcdURL
asset.dcTitle = 'The National Landcover Database 2006'
asset.dcPublisher = 'USGS'
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
"""@package GetGHCNDailyClimateDataForStationsInBoundingbox

@brief Query NCDC archive for climate data for all stations in the Global 
Historical Climatology Network
(http://www1.ncdc.noaa.gov/pub/data/ghcn/daily/readme.txt) that fall within a
bounding box. Will find all stations within the study area bounding box.  Requires that 
the  GHCN station database be setup using GHCNDSetup.py. Database must be 
stored in a location specified by a configuration file containing the section
'GHCND', and value 'PATH_OF_STATION_DB'.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GHCND', 'PATH_OF_STATION_DB'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   bbox_wgs84

Post conditions
---------------
1. Will write ClimatePointStation entries to the climate point section of metadata associated with the project directory:
   
2. Will save climate data files to outdir

Usage:
@code
GetGHCNDailyClimateDataForStationsInBoundingbox.py -p /path/to/project_dir
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 

@todo parse data files to determine start date, start time, and variables
"""
import os
import sys
import argparse
from datetime import datetime

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import ClimatePointStation
from ecohydrolib.spatialdata.utils import bboxFromString
#from ecohydrolib.spatialdata.utils import calculateBoundingBoxCenter
from ecohydrolib.climatedata.ghcndquery import findStationsWithinBoundingBox
from ecohydrolib.climatedata.ghcndquery import getClimateDataForStation

# Handle command line options
parser = argparse.ArgumentParser(description='Query NCDC archive for climate data for all Global Historical Climatology Network stations in the study area bounding box.')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-d', '--outdir', dest='outdir', required=False,
                    help='The name of the subdirectory within the project directory to write the climate data to.')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GHCND', 'PATH_OF_STATION_DB'):
    sys.exit("Config file %s does not define option %s in section %s" % \
          (configFile, 'PATH_OF_STATION_DB', 'GHCND'))

if args.outdir:
    outDir = args.outdir
else:
    outDir = 'climate'
outDirPath = os.path.join(context.projectDir, outDir)
if not os.path.exists(outDirPath):
    os.mkdir(outDirPath)

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
bbox = bboxFromString(studyArea['bbox_wgs84'])

# Find all GHCN stations within bounding box
stations = findStationsWithinBoundingBox(context.config, bbox)
print "Found %d stations in bounding box, downloading data..." % (len(stations))
# Get data for each station
for station in stations:
    outFile = os.path.join(outDir, station[0])
    returnCode = getClimateDataForStation(context.config, context.projectDir, outFile, station[0])
    assert(returnCode)
    
    # Write metadata
    newStation = ClimatePointStation()
    newStation.type = "GHCN"
    newStation.id = station[0]
    newStation.longitude = station[1]
    newStation.latitude = station[2]
    newStation.elevation = station[3]
    newStation.name = station[4]
    newStation.data = outFile
    #newStation.startDate = datetime.strptime("200001", "%Y%m")
    #newStation.endDate = datetime.strptime("200101", "%Y%m")
    #newStation.variables = [ClimatePointStation.VAR_TMIN, \
    #                    ClimatePointStation.VAR_TMAX]
    newStation.writeToMetadata(context)
    
# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
"""@package RegisterGage

@brief Register streamflow gage coordinates from a point shapefile into metadata 
store for a project directory, copying the shapefile into the project directory 
in the process. 

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   bbox_wgs84

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   gage [the name of the streamflow gage shapefile]  

2. Will write the following entry(ies) to the study area section of metadata associated with the project directory:
   gage_id_attr [the name of the attribute in the gage shapefile that uniquely identifies a streamflow gage]
   gage_id [the unique ID of the gage in the shapefile]

Usage:
@code
python ./RegisterGage.py -p /path/to/project_dir -g /path/to/gage/shapefile -l layername -a id_attribute -d id_value
@endcode

@note If option -t is not specified, UTM projection (WGS 84 coordinate system) will be inferred
from bounding box center.

@todo Set date in provenance to file modification date
"""
import os
import sys
import errno
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import deleteShapefile
from ecohydrolib.spatialdata.utils import getCoordinatesOfPointsFromShapefile
from ecohydrolib.spatialdata.utils import writeCoordinatePairsToPointShapefile
from ecohydrolib.spatialdata.utils import isCoordinatePairInBoundingBox


# Handle command line options
parser = argparse.ArgumentParser(description='Register streamflow gage shapefile with project')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-g', '--gageFile', dest='gageFile', required=True,
                    help='The name of the gage shapefile to be registered.')
parser.add_argument('-l', '--layerName', dest='layerName', required=True,
                    help='The name of the layer within the gage shapefile where gage points are located.')
parser.add_argument('-a', '--idAttribute', dest='idAttribute', required=True,
                    help='The name of the attribute field that uniquely identifies gage points.')
parser.add_argument('-d', '--idValue', dest='idValue', required=True,
                    help='The gage ID that uniquely identifies the gage point.')
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the gage shapefile to be written to the project directory.  File extension ".shp" will be added.')
parser.add_argument('-b', '--publisher', dest='publisher', required=False,
                    help="The publisher of the stream flow gage location dataset, if not supplied 'SELF PUBLISHED' will be used")
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not os.access(args.gageFile, os.R_OK):
    raise IOError(errno.EACCES, "Not allowed to read input gage shapefile %s" %
                  args.gageFile)
inGagePath = os.path.abspath(args.gageFile)

if args.publisher:
    publisher = args.publisher
else:
    publisher = 'SELF PUBLISHED'

if args.outfile:
    outfile = args.outfile
else:
    outfile = "gage"

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
bbox = bboxFromString(studyArea['bbox_wgs84'])

outFilename = "%s%sshp" % (outfile, os.extsep)
# Overwrite DEM if already present
outFilepath = os.path.join(context.projectDir, outFilename)
if os.path.exists(outFilepath):
    deleteShapefile(outFilepath)

sys.stdout.write('Importing streamflow gage...')
sys.stdout.flush()

gageIDs = [args.idValue]
coords = getCoordinatesOfPointsFromShapefile(inGagePath, args.layerName,
                                             args.idAttribute, gageIDs)
gage_lon = coords[0][0]
gage_lat = coords[0][1]
coordinates = (gage_lon, gage_lat)

# Ensure gage coordinates are within bounding box
if not isCoordinatePairInBoundingBox(bbox, coordinates):
    sys.exit("Gage coordinates %s, %s do not appear to lie within bounding box %s, %s, %s, %s" %
             ( str(gage_lon), str(gage_lat), str(bbox['minX']), str(bbox['minY']), str(bbox['maxX']), str(bbox['maxY']) ) )

shpFilename = writeCoordinatePairsToPointShapefile(context.projectDir, outfile, "gage_id", gageIDs, [coordinates])

sys.stdout.write('done\n')

# Write metadata
GenericMetadata.writeStudyAreaEntry(context, 'gage_id_attr', 'gage_id')
GenericMetadata.writeStudyAreaEntry(context, 'gage_id', args.idValue)
GenericMetadata.writeStudyAreaEntry(context, 'gage_lat_wgs84', gage_lat)
GenericMetadata.writeStudyAreaEntry(context, 'gage_lon_wgs84', gage_lon)

# Write provenance
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'gage'
asset.dcIdentifier = shpFilename
asset.dcSource = "file://%s" % (inGagePath,)
asset.dcTitle = 'Streamflow gage'
asset.dcPublisher = publisher
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package GetSSURGOFeaturesForBoundingbox

@brief Query USDA soil datamart for SSURGO MapunitPolyExtended features and the following 
attributes, computed as a weighted average of the components in each mapunit, for the first 
soil horizon of each mapunit: ksat (chorizon.ksat_r), pctClay (chorizon.claytotal_r), 
pctSilt (chorizon.silttotal_r), pctSand (chorizon.andtotal_r), porosity (chorizon.wsatiated_r), 
'fieldCap' (chorizon.wthirdbar_r), 'avlWatCap' (plant available water capacity; chorizon.awc_r), 
and drnWatCont (drainable water capacity; porosity - fieldCap).  

@note For information on SSURGO attributes see: http://soildatamart.nrcs.usda.gov/SSURGOMetadata.aspx

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  

Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GDAL/OGR', 'PATH_OF_OGR2OGR'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   bbox_wgs84
   dem_res_x
   dem_res_y
   dem_srs

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   soil_features [the name of the vector file containing the soil features]

Usage:
@code
GetSSURGOFeaturesForBoundingbox.py -p /path/to/project_dir
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified.
"""
import os
import sys
import argparse
import textwrap

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance

from ecohydrolib.spatialdata.utils import deleteShapefile
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import convertGMLToShapefile
from ecohydrolib.ssurgo.featurequery import getMapunitFeaturesForBoundingBox
from ecohydrolib.ssurgo import featurequery
   
# Handle command line options
parser = argparse.ArgumentParser(description='Get SSURGO features for a bounding box')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing SSURGO features shapefile in project directory.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GDAL/OGR', 'PATH_OF_OGR2OGR'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_OGR2OGR'))

# Check if features shapefile already exists
manifest = GenericMetadata.readManifestEntries(context)
if 'soil_features' in manifest:
    if args.overwrite:
        sys.stdout.write('Deleting existing SSURGO features shapefile\n')
        sys.stdout.flush()
        shpFilepath = os.path.join( context.projectDir, manifest['soil_features'] )
        deleteShapefile(shpFilepath)
    else:
        sys.exit( textwrap.fill('SSURGO features already exist in project directory.  Use --overwrite option to overwrite.') )

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
bbox = bboxFromString(studyArea['bbox_wgs84'])

outputrasterresolutionX = studyArea['dem_res_x']
outputrasterresolutionY = studyArea['dem_res_y']
srs = studyArea['dem_srs']

sys.stdout.write('Downloading SSURGO features for study area from USDA Soil Data mart...\n')
sys.stdout.flush()
shpFilename = getMapunitFeaturesForBoundingBox(context.config, context.projectDir, bbox, t_srs=srs)[0]

# Write provenance
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'soil_features'
asset.dcIdentifier = shpFilename
asset.dcSource = featurequery.WFS_URL
asset.dcTitle = 'SSURGO soils data'
asset.dcPublisher = 'USDA'
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
"""@package GenerateSoilPropertyRastersFromSOLIM

@brief Infer soil attributes associated with SSURGO mapunit features using SOLIM

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  

Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'SOLIM', 'PATH_OF_SOLIM'

2. The following metadata entry(ies) must be present in the manifest section of the metadata associated with the project directory:
   soil_features [the name of the vector file containing the soil features]

3. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   dem_res_x
   dem_res_y

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   soil_raster_<attr> [the name of the raster file for each soil property raster]

Usage:
@code
GenerateSoilPropertyRastersFromSOLIM.py -p /path/to/project_dir
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified.

@todo Recognize when DEM resolution is less than 10m, resample temporary DEM, run SOLIM with resampled DEM
"""
import os
import sys
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.ssurgo import rasterize
from ecohydrolib.ssurgo import attributequery
from ecohydrolib.solim.inference import inferSoilPropertiesForSSURGOAndTerrainData     

# Handle command line options
parser = argparse.ArgumentParser(description='Get SSURGO features for a bounding box')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('SOLIM', 'PATH_OF_SOLIM'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'SOLIM', 'PATH_OF_SOLIM'))

# Get provenance data for SSURGO
ssurgoProvenance = [i for i in GenericMetadata.readAssetProvenanceObjects(context) if i.name == 'soil_features'][0]
if ssurgoProvenance is None:
    sys.exit("Unable to load SSURGO provenance information from metadata")

# Get manifest entries
manifest = GenericMetadata.readManifestEntries(context)
shpFilename = manifest['soil_features']
shpFilepath = os.path.join(context.projectDir, shpFilename)
demFilename = manifest['dem']
demFilepath = os.path.join(context.projectDir, demFilename)
layerName = os.path.splitext(shpFilename)[0]

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
outputrasterresolutionX = studyArea['dem_res_x']
outputrasterresolutionY = studyArea['dem_res_y']

# Truncate attributes to 10 characters because shapefiles rely on ancient technology
sys.stdout.write('Generating soil property maps using SOLIM...')
sys.stdout.flush()
attrList = [elem[:10] for elem in attributequery.ATTRIBUTE_LIST_NUMERIC] 
rasterFiles = inferSoilPropertiesForSSURGOAndTerrainData(config=context.config, outputDir=context.projectDir, \
                                                         shpFilepath=shpFilepath, demFilepath=demFilepath, \
                                                         featureAttrList=attrList)
sys.stdout.write('done\n')

# Write metadata entries
for attr in rasterFiles.keys():
    asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
    asset.name = "soil_raster_%s" % (attr,)
    asset.dcIdentifier = rasterFiles[attr]
    asset.dcSource = 'http://solim.geography.wisc.edu'
    asset.dcTitle = attr
    asset.dcPublisher = 'Department of Geography, UW-Madison'
    asset.dcDescription = cmdline
    asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)


#!/usr/bin/env python
"""@package GetHYDRO1kDEMForBoundingbox

@brief Extract tile for HYDRO1k digital elevation model (DEM) stored locally

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GDAL/OGR', 'PATH_OF_GDAL_WARP'
   'HYDRO1k', 'PATH_OF_HYDRO1K_DEM'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   bbox_wgs84

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   dem [the name of the DEM raster]  

2. Will write the following entry(ies) to the study area section of metadata associated with the project directory:
   dem_res_x [X resolution of the DEM raster in units of the raster's projection]
   dem_res_y [Y resolution of the DEM raster in units of the raster's projection]
   dem_srs [spatial reference system of the DEM, in EPSG:<nnnn> format]
   dem_columns [number of pixels in the X direction]
   dem_rows [number of pixels in the Y direction]

Usage:
@code
GetHYDRO1kDEMForBoundingbox.py -p /path/to/project_dir -s 3 3
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 

@note If option -t is not specified, UTM projection (WGS 84 coordinate system) will be inferred
from bounding box center.

@todo Debug case where t_srs is specified as NAD83, with unit degree, and 1000 is used as default DEM resolution
"""
import os
import sys
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.hydro1k import demtile
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import resampleRaster
from ecohydrolib.spatialdata.utils import getSpatialReferenceForRaster
from ecohydrolib.spatialdata.utils import getDimensionsForRaster
from ecohydrolib.spatialdata.utils import deleteGeoTiff
from ecohydrolib.spatialdata.utils import calculateBoundingBoxCenter
from ecohydrolib.spatialdata.utils import getUTMZoneFromCoordinates
from ecohydrolib.spatialdata.utils import getEPSGStringForUTMZone

# Handle command line options
parser = argparse.ArgumentParser(description='Get DEM raster (in GeoTIFF format) for a bounding box from GeoBrain WCS4DEM')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the DEM file to be written.  File extension ".tif" will be added.')
parser.add_argument('-s', '--demResolution', dest='demResolution', required=False, nargs=2, type=float,
                    help='Two floating point numbers representing the desired X and Y output resolution of soil property raster maps; unit: meters if SRS is UTM')
parser.add_argument('-t', '--t_srs', dest='t_srs', required=False, 
                    help='Target spatial reference system of output, in EPSG:num format')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GDAL/OGR', 'PATH_OF_GDAL_WARP'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_GDAL_WARP'))

if args.outfile:
    outfile = args.outfile
else:
    outfile = "DEM"

demFilename = "%s.tif" % (outfile)
# Overwrite DEM if already present
demFilepath = os.path.join(context.projectDir, demFilename)
if os.path.exists(demFilepath):
    deleteGeoTiff(demFilepath)

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
bbox = bboxFromString(studyArea['bbox_wgs84'])

# Determine target spatial reference
if args.t_srs:
    t_srs = args.t_srs
else:
    # Default for UTM
    (centerLon, centerLat) = calculateBoundingBoxCenter(bbox)
    (utmZone, isNorth) = getUTMZoneFromCoordinates(centerLon, centerLat)
    t_srs = getEPSGStringForUTMZone(utmZone, isNorth)

# Get DEM from HYDRO1k dataset on disk
tmpDEMFilename = "%s-TEMP.tif" % (outfile)
returnCode = demtile.getDEMForBoundingBox(context.config, context.projectDir, tmpDEMFilename, bbox=bbox, srs=t_srs)
assert(returnCode)
tmpDEMFilepath = os.path.join(context.projectDir, tmpDEMFilename)

if args.demResolution:
    demResolutionX = args.demResolution[0]
    demResolutionY = args.demResolution[1]
else:
    demSrs = getSpatialReferenceForRaster(tmpDEMFilepath)
    demResolutionX = demSrs[0]
    demResolutionY = demSrs[1]
    
# Resample DEM to target srs and resolution
resampleRaster(context.config, context.projectDir, tmpDEMFilepath, demFilename, \
               s_srs=demtile.DEFAULT_CRS, t_srs=t_srs, \
               trX=demResolutionX, trY=demResolutionY)

# Write metadata
GenericMetadata.writeStudyAreaEntry(context, "dem_res_x", demResolutionX)
GenericMetadata.writeStudyAreaEntry(context, "dem_res_y", demResolutionY)
GenericMetadata.writeStudyAreaEntry(context, "dem_srs", t_srs)

# Get rows and columns for DEM
(columns, rows) = getDimensionsForRaster(demFilepath)
GenericMetadata.writeStudyAreaEntry(context, "dem_columns", columns)
GenericMetadata.writeStudyAreaEntry(context, "dem_rows", rows)

# Write provenance
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'dem'
asset.dcIdentifier = demFilename
asset.dcSource = 'http://eros.usgs.gov/#/Find_Data/Products_and_Data_Available/gtopo30/hydro/namerica'
asset.dcTitle = 'Digital Elevation Model from HYDRO1k'
asset.dcPublisher = 'USGS'
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)

# Clean-up
deleteGeoTiff(tmpDEMFilepath)


#!/usr/bin/env python
"""@package RegisterDEM

@brief Register digital elevation model (DEM) data into metadata store for a project directory,
copying the DEM file into the project directory in the process.  Will create a study area 
polygon shapefile for the extent of the DEM imported. 

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
None

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   dem [the name of the DEM raster]
   study_area_shapefile [the name of the study area shapefile generated from the DEM raster extent]

2. Will write the following entry(ies) to the study area section of metadata associated with the project directory:
   bbox_wgs84
   dem_res_x [X resolution of the DEM raster in units of the raster's projection]
   dem_res_y [Y resolution of the DEM raster in units of the raster's projection]
   dem_srs [spatial reference system of the DEM, in EPSG:<nnnn> format]
   dem_columns [number of pixels in the X direction]
   dem_rows [number of pixels in the Y direction]

Usage:
@code
python ./RegisterDEM.py -p /path/to/project_dir -d /demfile/to/register
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 

@todo Add targer SRS, if present, call GDAL warp instead of just copying the file
@todo Set date in provenance to file modification date
"""
import os
import sys
import errno
import argparse
import textwrap

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.spatialdata.utils import isValidSrs
from ecohydrolib.spatialdata.utils import RASTER_RESAMPLE_METHOD
from ecohydrolib.spatialdata.utils import copyRasterToGeoTIFF
from ecohydrolib.spatialdata.utils import resampleRaster
from ecohydrolib.spatialdata.utils import getDimensionsForRaster
from ecohydrolib.spatialdata.utils import getSpatialReferenceForRaster
from ecohydrolib.spatialdata.utils import getBoundingBoxForRaster
from ecohydrolib.spatialdata.utils import writeBboxPolygonToShapefile

# Handle command line options
parser = argparse.ArgumentParser(description='Register DEM with project')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-d', '--demfile', dest='demfile', required=True,
                    help='The name of the DEM file to be registered.')
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the DEM file to be written to the project directory.  File extension ".tif" will be added.')
parser.add_argument('-b', '--publisher', dest='publisher', required=False,
                    help="The publisher of the DEM dataset, if not supplied 'SELF PUBLISHED' will be used")
parser.add_argument('-c', '--demResolution', dest='demResolution', required=False, nargs=2, type=float,
                    help='Two floating point numbers representing the desired X and Y output resolution of soil property raster maps; unit: meters')
parser.add_argument('-t', '--t_srs', dest='t_srs', required=False, 
                    help='Target spatial reference system of output, in EPSG:num format')
parser.add_argument('-s', '--resampleMethod', dest='resampleMethod', required=False,
                    choices=RASTER_RESAMPLE_METHOD, default='bilinear',
                    help='Method to use to resample DEM (if necessary). Defaults to bilinear.')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GDAL/OGR', 'PATH_OF_GDAL_WARP'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_GDAL_WARP'))

if not os.access(args.demfile, os.R_OK):
    raise IOError(errno.EACCES, "Not allowed to read input DEM %s" (args.demfile,))
inDEMPath = os.path.abspath(args.demfile)
inSpatialMetadata = getSpatialReferenceForRaster(inDEMPath)

resample = False

s_srs = inSpatialMetadata[5]
t_srs = s_srs
if args.t_srs:
    if not isValidSrs(args.t_srs):
        sys.exit(textwrap.fill("ERROR: '%s' is not a valid spatial reference.  Spatial reference must be of the form 'EPSG:XXXX', e.g. 'EPSG:32617'.  For more information, see: http://www.spatialreference.org/" % (args.t_srs,) ) )
    if s_srs != args.t_srs:
        t_srs = args.t_srs
        resample = True

if args.demResolution:
    if args.demResolution[0] != inSpatialMetadata[0] or args.demResolution[1] != inSpatialMetadata[1]:
        resample = True
    demResolutionX = args.demResolution[0]
    demResolutionY = args.demResolution[1]
else:
    demResolutionX = inSpatialMetadata[0]
    demResolutionY = inSpatialMetadata[1]

if args.publisher:
    publisher = args.publisher
else:
    publisher = 'SELF PUBLISHED'

if args.outfile:
    outfile = args.outfile
else:
    outfile = "DEM"

demFilename = "%s%stif" % (outfile, os.extsep)
# Overwrite DEM if already present
demFilepath = os.path.join(context.projectDir, demFilename)
if os.path.exists(demFilepath):
    os.unlink(demFilepath)

# Copy the raster in to the project directory (reprojecting if need be)
if not resample:
    sys.stdout.write("Importing DEM...")
    sys.stdout.flush()
    copyRasterToGeoTIFF(context.config, context.projectDir, inDEMPath, demFilename)
else:
    sys.stdout.write("Reprojecting DEM from %s to %s, spatial resolution (%.2f, %.2f) to (%.2f, %.2f)..." % \
                     (s_srs, t_srs, inSpatialMetadata[0], inSpatialMetadata[1],
                      demResolutionX, demResolutionY) )
    sys.stdout.flush()
    resampleRaster(context.config, context.projectDir, inDEMPath, demFilename,
                   s_srs, t_srs, demResolutionX, demResolutionY, args.resampleMethod)
sys.stdout.write('done\n')
# Get the bounding box for the DEM
bbox = getBoundingBoxForRaster(demFilepath)
# Write a shapefile for the bounding box
shpFilename = writeBboxPolygonToShapefile(bbox, context.projectDir, "studyarea")

# Write metadata
GenericMetadata.writeStudyAreaEntry(context, "bbox_wgs84", "%f %f %f %f" % (bbox['minX'], bbox['minY'], bbox['maxX'], bbox['maxY']))

# Get spatial metadata for DEM
demSpatialMetadata = getSpatialReferenceForRaster(demFilepath)
GenericMetadata.writeStudyAreaEntry(context, "dem_res_x", demSpatialMetadata[0])
GenericMetadata.writeStudyAreaEntry(context, "dem_res_y", demSpatialMetadata[1])
GenericMetadata.writeStudyAreaEntry(context, "dem_srs", demSpatialMetadata[5])

# Get rows and columns for DEM
demFilepath = os.path.join(context.projectDir, demFilename)
(columns, rows) = getDimensionsForRaster(demFilepath)
GenericMetadata.writeStudyAreaEntry(context, "dem_columns", columns)
GenericMetadata.writeStudyAreaEntry(context, "dem_rows", rows)

# Write provenance
inputDEMURL = "file://%s" % (inDEMPath,)
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'dem'
asset.dcIdentifier = demFilename
asset.dcSource = inputDEMURL
asset.dcTitle = 'Digital Elevation Model'
asset.dcPublisher = publisher
asset.dcDescription = cmdline
asset.writeToMetadata(context)

asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'study_area_shapefile'
asset.dcIdentifier = shpFilename
asset.dcSource = inputDEMURL
asset.dcTitle = 'Study area shapefile'
asset.dcPublisher = publisher
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)



"""@package ecohydrolib.ssurgo.featurequery
    
@brief Make feature queries against USDA Soil Data Mart OGC web service interface

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import sys
import errno
import xml.sax
import json

from owslib.wfs import WebFeatureService

from ecohydrolib.spatialdata.utils import calculateBoundingBoxArea
from ecohydrolib.spatialdata.utils import tileBoundingBox
from ecohydrolib.spatialdata.utils import convertGMLToGeoJSON
from ecohydrolib.spatialdata.utils import convertGeoJSONToShapefile
from attributequery import getParentMatKsatTexturePercentClaySiltSandForComponentsInMUKEYs
from attributequery import joinSSURGOAttributesToFeaturesByMUKEY_GeoJSON
from attributequery import computeWeightedAverageKsatClaySandSilt
from saxhandlers import SSURGOFeatureHandler       

MAX_SSURGO_EXTENT = 10000000000 # 10,100,000,000 sq. meters
MAX_SSURGO_EXTENT = MAX_SSURGO_EXTENT / 10

WFS_URL = 'http://SDMDataAccess.nrcs.usda.gov/Spatial/SDMWGS84Geographic.wfs'

def getMapunitFeaturesForBoundingBox(config, outputDir, bbox, tileBbox=False, t_srs='EPSG:4326'):
    """ Query USDA Soil Data Mart for SSURGO MapunitPolyExtended features with a given bounding box.
        Features will be written to one or more shapefiles, one file for each bboxTile tile,
        stored in the specified output directory. The filename will be returned as a string.
        Will fetch SSURGO tabular data (see ssurgolib.attributequery.ATTRIBUTE_LIST for a list
        of attributes) and join those data to the features in the final shapefiles(s).
    
        @note Will silently exit if features already exist.
    
        @param config onfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_OGR2OGR'
        @param outputDir String representing the absolute/relative path of the directory into which features should be written
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        @param tileBoundingBox True if bounding box should be tiled if extent exceeds featurequery.MAX_SSURGO_EXTENT
        @param t_srs String representing the spatial reference system of the output shapefiles, of the form 'EPSG:XXXX'
        
        @return A list of strings representing the name of the shapefile(s) to which the mapunit features were saved.
        
        @exception IOError if output directory is not a directory
        @exception IOError if output directory is not writable
        @exception Exception if bounding box area is greater than MAX_SSURGO_EXTENT
        @exception Exception if no MUKEYs were returned
    """
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)

    typeName = 'MapunitPolyExtended'

    if tileBbox:
        bboxes = tileBoundingBox(bbox, MAX_SSURGO_EXTENT)
        sys.stderr.write("Dividing bounding box %s into %d tiles\n" % (str(bbox), len(bboxes)))
    else:
        if calculateBoundingBoxArea(bbox, t_srs) > MAX_SSURGO_EXTENT:
            raise Exception("Bounding box area is greater than %f sq. meters" % (MAX_SSURGO_EXTENT,))
        bboxes = [bbox]
    
    outFiles = []
    
    for bboxTile in bboxes:
        minX = bboxTile['minX']; minY = bboxTile['minY']; maxX = bboxTile['maxX']; maxY = bboxTile['maxY']
        bboxLabel = str(minX) + "_" + str(minY) + "_" + str(maxX) + "_" + str(maxY)
    
        gmlFilename = "%s_bbox_%s-attr.gml" % (typeName, bboxLabel)
        gmlFilepath = os.path.join(outputDir, gmlFilename)
    
        if not os.path.exists(gmlFilepath):
            sys.stderr.write("Fetching SSURGO data for sub bboxTile %s\n" % bboxLabel)
        
            wfs = WebFeatureService(WFS_URL, version='1.0.0')
            filter = "<Filter><BBOX><PropertyName>Geometry</PropertyName> <Box srsName='EPSG:4326'><coordinates>%f,%f %f,%f</coordinates> </Box></BBOX></Filter>" % (minX, minY, maxX, maxY)
            gml = wfs.getfeature(typename=(typeName,), filter=filter, propertyname=None)
    
            # Write intermediate GML to a file
            intGmlFilename = "%s_bbox_%s.gml" % (typeName, bboxLabel)
            intGmlFilepath = os.path.join(outputDir, intGmlFilename)
            out = open(intGmlFilepath, 'w')
            out.write(gml.read())
            out.close()
            
            # Parse GML to get list of MUKEYs
            gmlFile = open(intGmlFilepath, 'r')
            ssurgoFeatureHandler = SSURGOFeatureHandler()
            xml.sax.parse(gmlFile, ssurgoFeatureHandler)
            gmlFile.close()
            mukeys = ssurgoFeatureHandler.mukeys
            
            if len(mukeys) < 1:
                raise Exception("No SSURGO features returned from WFS query.  SSURGO GML format may have changed.\nPlease contact the developer.")
            
            # Get attributes (ksat, texture, %clay, %silt, and %sand) for all components in MUKEYS
            attributes = getParentMatKsatTexturePercentClaySiltSandForComponentsInMUKEYs(mukeys)
            
            # Compute weighted average of soil properties across all components in each map unit
            avgAttributes = computeWeightedAverageKsatClaySandSilt(attributes)
            
            # Convert GML to GeoJSON so that we can add fields easily (GDAL 1.10+ validates GML schema 
            #   and won't let us add fields)
            tmpGeoJSONFilename = convertGMLToGeoJSON(config, outputDir, intGmlFilepath, typeName)
            tmpGeoJSONFilepath = os.path.join(outputDir, tmpGeoJSONFilename)
            
            # Join map unit component-averaged soil properties to attribute table in GML file
#             gmlFile = open(intGmlFilepath, 'r')
#             joinedGmlStr = joinSSURGOAttributesToFeaturesByMUKEY(gmlFile, typeName, avgAttributes)
#             gmlFile.close()
            tmpGeoJSONFile = open(tmpGeoJSONFilepath, 'r')
            geojson = json.load(tmpGeoJSONFile)
            tmpGeoJSONFile.close()
            joinSSURGOAttributesToFeaturesByMUKEY_GeoJSON(geojson, typeName, avgAttributes)
            
            # Write Joined GeoJSON to a file
            out = open(tmpGeoJSONFilepath, 'w')
            json.dump(geojson, out)
            out.close()
            
            # Convert GeoJSON to shapefile
            filename = os.path.splitext(intGmlFilename)[0]
            shpFilename = convertGeoJSONToShapefile(config, outputDir, tmpGeoJSONFilepath, filename, t_srs=t_srs)
            
            # Delete intermediate files
            os.unlink(intGmlFilepath)
            os.unlink(tmpGeoJSONFilepath)
        
        outFiles.append(shpFilename)
    
    # TODO: join tiled data if tileBbox
        
    return outFiles
    
    


"""@package ecohydrolib.ssurgo.saxhandlers
    
@brief xml.sax.ContentHandler subclass for parsing SSURGO features stored as GML files

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import xml.sax

from oset import oset

class SSURGOFeatureHandler(xml.sax.ContentHandler):
    """ Parse SSURGO features fetched from USDA soil datamart WFS 1.0.0 web service.
        Currently pulls out MUKEYs for each feature returned as either MapunitPoly or MapunitPolyExtended
        For example:
            /wfs:FeatureCollection/gml:featureMember/ms:MapunitPolyExtended/ms:MUKEY
            /wfs:FeatureCollection/gml:featureMember/ms:MapunitPoly/ms:MUKEY
        @note Parses with namespaces turned off. Parse also assumes XML is well-formed
    """
    wfs_FeatureCollection = "wfs:FeatureCollection"
    gml_featureMember = "gml:featureMember"
    ms_MapunitPoly = "ms:MapunitPoly"
    ms_MapunitPolyExtended = "ms:MapunitPolyExtended"
    ms_MUKEY = "ms:mukey"
    
    inWfsFeatureCollection = False
    inGmlFeatureMember = False
    inMapunitPolyExtended = False
    inMapunitPoly = False
    getMukey = False
    
    mukeys = None
    
    def __init__(self):
        """ Default constructor
            
            @param self This object
        """
        xml.sax.ContentHandler.__init__(self)
        self.mukeys = list()
        
    def characters(self, content):
        """ Called when character data are encountered
        
            @param self This object
            @param content String containing characters
        """
        if self.getMukey:
            #print "Saving MUKEY %s" % (content,)
            self.mukeys.append(content)
            
    def startElement(self, name, attrs):
        """ Called at the start of an element
        
            @param self This object
            @param name The name of the object
            @param attrs Attributes of the element
        """
        #print "startElement: " + name
        if (self.inMapunitPolyExtended or self.inMapunitPoly) and ( self.ms_MUKEY == name.lower() ):
            self.getMukey = True
        elif self.inGmlFeatureMember and (self.ms_MapunitPolyExtended == name):
            self.inMapunitPolyExtended = True
        elif self.inGmlFeatureMember and (self.ms_MapunitPoly == name):
            self.inMapunitPoly = True
        elif self.inWfsFeatureCollection and (self.gml_featureMember == name):
            self.inGmlFeatureMember = True
        elif self.wfs_FeatureCollection == name:
            self.inWfsFeatureCollection = True
            
            
    def endElement(self, name):
        """ Called at the end of an element
        
            @param self This object
            @param name The name of the object
        """
        if self.getMukey and (self.ms_MUKEY == name):
            self.getMukey = False
        elif self.inMapunitPolyExtended and (self.ms_MapunitPolyExtended == name):
            self.inMapunitPolyExtended = False
        elif self.inMapunitPoly and (self.ms_MapunitPoly == name):
            self.inMapunitPoly = False
        elif self.inGmlFeatureMember and (self.gml_featureMember == name):
            self.inGmlFeatureMember = False
        elif self.inWfsFeatureCollection and (self.wfs_FeatureCollection == name):
            self.inWfsFeatureCollection = False
            

class SSURGOMUKEYQueryHandler(xml.sax.ContentHandler):
    """ Parse query results fetched via USDA soil datamart tabular query web service.
        Query results are assumed to contain one or more attributes associated with MUKEYs.
        Supports results who contain multiple "rows" for a given MUKEY
    
        Stores results in a list of lists named "results". Each row is stored in the outer 
        list, with column values being stored in the inner list.
    
        Store column names in an ordered set (oset.oset) named "columnNames".
    
        @note Parses with namespaces turned off. Parse also assumes XML is well-formed.
          Assumes that the order of column names seen in the first row is
          the same order used for all rows. If this is not the case, columnNames
          will not represent the column names for all data in results.
    """
    NewDataSet = "NewDataSet"
    Table = "Table"
    mukey = "mukey"
    
    inNewDataSet = False
    inTable = False
    recordColData = False
    elementStack = None
    
    columnNames = None
    results = None
    
    _wroteResults = False
    _tmpColData = None
    
    def __init__(self):
        xml.sax.ContentHandler.__init__(self)
        self.results = list()
        self.elementStack = list()
        self.columnNames = oset()
        
    def characters(self, content):
        if self.recordColData:
            self.columnNames.add(self.elementStack[len(self.elementStack)-1])
            self._tmpColData.append(content) 
            self._wroteResults = True   
        
    def startElement(self, name, attrs):
        self.elementStack.append(name)
        
        if self.inTable:
            self.recordColData = True
        elif self.inNewDataSet and (self.Table == name):
            self.inTable = True
            self._tmpColData = list()
        elif self.NewDataSet == name:
            self.inNewDataSet = True
    
    def endElement(self, name):
        
        if self.recordColData and (self.elementStack[len(self.elementStack)-1] == name):
            if not self._wroteResults:
                # Handle case where attribute element was empty: make sure we write a blank string for the value
                self._tmpColData.append('')
            self.recordColData = False
            self._wroteResults = False
        elif self.inTable and (self.Table == name):
            self.inTable = False
            # Store column data
            self.results.append(self._tmpColData)
            self._tmpColData = None
        elif self.inNewDataSet and (self.NewDataSet == name):
            self.inNewDataSet = False
        
        self.elementStack.pop()
    


#!/usr/bin/env python
"""@package NHDPlusV2Setup
    
@brief Builds SQLite3 databases from NHDPlus V2 data archives downloaded from
http://www.horizon-systems.com/NHDPlus/NHDPlusV2_home.php
@brief Assembles regional NHDPlus V2 data into national dataset suitable for gage-
reach based region of interest extraction

@note Requires GDAL/OGR that has been built with SQLite3 support
@note Requires 7z to extract data from archive files

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

Usage:
@code
NHDPlusSetup.py -i <config_file> -a  <archive_dir> -o <output_dir>
@endcode
"""
import os
import sys
import errno
import argparse
import subprocess
import sqlite3
import ConfigParser

from ecohydrolib.dbf import dbfreader


parser = argparse.ArgumentParser(description='Assemble regional NHDPLus V2 data into a national dataset')
parser.add_argument('-i', '--configfile', dest='configfile', required=True,
                    help='The configuration file')
parser.add_argument('-a', '--archive', dest='archiveDir', 
                    required=True,
                    help='Directory in which to look for NHDPLus V2 .7z archives')
parser.add_argument('-o', '--output', dest='outputDir', 
                    required=True,
                    help='Directory to which processed NHDPLus V2 data should be placed')
parser.add_argument('-s1', '--skipUnzip', dest='skipUnzip', action='store_true',
                    default=False, required=False,
                    help='Skip step where archives are unzipped to output directory')
parser.add_argument('-s2', '--skipCatchment', dest='skipCatchment', action='store_true',
                    default=False, required=False,
                    help='Skip step where regional catchment shapefiles are merged')
parser.add_argument('-s3', '--skipDB', dest='skipDB', action='store_true',
                    default=False, required=False,
                    help='Skip step where database is created')
parser.add_argument('-s4', '--skipGageLoc', dest='skipGageLoc', action='store_true',
                    default=False, required=False,
                    help='Skip step where GageLoc database is created')
args = parser.parse_args()

config = ConfigParser.RawConfigParser()
config.read(args.configfile)

if not config.has_option('GDAL/OGR', 'PATH_OF_OGR2OGR'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_OGR2OGR'))
if not config.has_option('UTIL', 'PATH_OF_FIND'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'UTIL', 'PATH_OF_FIND'))
if not config.has_option('UTIL', 'PATH_OF_SEVEN_ZIP'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'UTIL', 'PATH_OF_SEVEN_ZIP'))
if not config.has_option('UTIL', 'PATH_OF_SQLITE'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'UTIL', 'PATH_OF_SQLITE'))
    
pathOfOgr = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
pathOfFind = config.get('UTIL', 'PATH_OF_FIND')
pathOfSevenZip = config.get('UTIL', 'PATH_OF_SEVEN_ZIP')
pathOfSqlite = config.get('UTIL', 'PATH_OF_SQLITE')

if not os.access(args.archiveDir, os.R_OK):
    raise IOError(errno.EACCES, "Not allowed to read from archive directory %s" %
                  args.archiveDir)

if not os.access(args.outputDir, os.W_OK):
    raise IOError(errno.EACCES, "Not allowed to write to output directory %s" %
                  args.outputDir)

nhdPlusDB = os.path.join(args.outputDir, "NHDPlusDB.sqlite")

# 0. Unpacking NHDPlus archives into output directory
if not args.skipUnzip:
    print("Unpacking NHDPlus archives into output directory %s" % (args.outputDir,))

    # Get a list of zip files
    zipFiles = subprocess.check_output("%s %s -type f -name *.7z -print" % (pathOfFind, args.archiveDir,), shell=True).split()

    # Unpack zip files into output directory
    for file in zipFiles:
        sevenZCommand = "%s x -y -o%s %s" % \
            (pathOfSevenZip, args.outputDir, file)
        print sevenZCommand
        returnCode = os.system(sevenZCommand)
        assert(returnCode == 0)

# 1. Find GageLoc shapefile and convert it to a spatial SQLite DB
if not args.skipGageLoc:
    print("Converting GageLoc shapefile to sqlite database ...")
    gageLocDB = os.path.join(args.outputDir, "GageLoc.sqlite")
    gageLoc = subprocess.check_output("%s %s -type f -iname GageLoc.shp -print" % (pathOfFind, args.outputDir,), shell=True).split()
    assert(gageLoc)
    gageLocShp = gageLoc[0]
    assert(os.access(gageLocShp, os.R_OK))
    ogrCommand = '%s -gt 65536 -f "SQLite" -t_srs "EPSG:4326" %s %s' % (pathOfFind, gageLocDB, gageLocShp)
    returnCode = os.system(ogrCommand)
    assert(returnCode == 0) 
    
    # Index fields
    sqliteCommand = "%s %s 'CREATE INDEX IF NOT EXISTS reachcode_measure_idx on GageLoc (reachcode,measure)'" % (pathOfSqlite, gageLocDB)
    returnCode = os.system(sqliteCommand)
    assert(returnCode == 0)
    
    sqliteCommand = "%s %s 'CREATE INDEX IF NOT EXISTS gage_loc_source_fea_idx ON GageLoc (source_fea)'" % (pathOfSqlite, gageLocDB)
    returnCode = os.system(sqliteCommand)
    assert(returnCode == 0)
    #cursor.execute("""""")
    

# 2. Find catchment shapefiles
if not args.skipCatchment:
    conusCatchment = os.path.join(args.outputDir, "Catchment.sqlite")
    #print conusCatchment

    # Remove existing conusCatchment
    if os.access(conusCatchment, os.F_OK):
        os.remove(conusCatchment)

    print("Finding catchment shapefiles")
    shapefiles = subprocess.check_output("%s %s -type f -iname Catchment.shp -print" % (pathOfFind, args.outputDir,), shell=True).split()
    #print shapefiles

    # 3. Intersect all catchment shapefiles into one shapefile for the entire CONUS
    print("Intersecting regional catchment shapefiles in to single CONUS catchment feature dataset ...")
    numFiles = len(shapefiles)
    currFile = 0
    for file in shapefiles:
        pctComplete = (float(currFile) / float(numFiles)) * 100
        currFile = currFile + 1
        ogrCommand = '%s -gt 65536 -f "SQLite" -append %s %s' % (pathOfOgr, conusCatchment, file)
        #print ogrCommand
        sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)" % (currFile, numFiles, pctComplete))
        sys.stdout.flush()
        returnCode = os.system(ogrCommand)
        assert(returnCode == 0)

    pctComplete = (float(currFile) / float(numFiles)) * 100
    sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)\n" % (currFile, numFiles, pctComplete))

    # 4. Add index to CONUS catchment
    print "Indexing CONUS shapefile (this may take a while) ..."
    sqliteCommand = "%s %s 'CREATE INDEX IF NOT EXISTS featureid_idx on catchment (featureid)'" % (pathOfSqlite, conusCatchment)
    returnCode = os.system(sqliteCommand)
    assert(returnCode == 0)

# 5. Create NHDPlus SQLite database to store flowline and stream gage records
if not args.skipDB:
    # Remove existing database if it's there
    if os.access(nhdPlusDB, os.F_OK):
        os.remove(nhdPlusDB)
    #conn = sqlite3.connect(nhdPlusDB, isolation_level=None)
    conn = sqlite3.connect(nhdPlusDB)
    cursor = conn.cursor()
    
    # Create PlusFlowlineVAA table and indices
    cursor.execute("""CREATE TABLE IF NOT EXISTS PlusFlowlineVAA
    (ComID INTEGER,
    Fdate DATETIME,
    StreamLeve INTEGER,
    StreamOrde INTEGER,
    StreamCalc INTEGER,
    FromNode INTEGER,
    ToNode INTEGER,
    Hydroseq INTEGER,
    LevelPathI INTEGER,
    Pathlength REAL,
    TerminalPa INTEGER,
    ArbolateSu REAL,
    Divergence INTEGER,
    StartFlag INTEGER,
    TerminalFl INTEGER,
    DnLevel INTEGER,
    ThinnerCod INTEGER,
    UpLevelPat INTEGER,
    UpHydroseq INTEGER,
    DnLevelPat INTEGER,
    DnMinorHyd INTEGER,
    DnDrainCou INTEGER,
    DnHydroseq INTEGER,
    FromMeas REAL,
    ToMeas REAL,
    ReachCode TEXT,
    LengthKM REAL,
    Fcode INTEGER,
    RtnDiv INTEGER,
    OutDiv INTEGER,
    DivEffect INTEGER,
    VPUIn INTEGER,
    VPUOut INTEGER,
    TravTime INTEGER,
    PathTime INTEGER,
    AreaSqKM REAL,
    TotDASqKM REAL,
    DivDASqKM REAL)
    """)
    cursor.execute("""CREATE INDEX IF NOT EXISTS PlusFlowlineVAA_Comid_idx ON PlusFlowlineVAA (ComID)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS PlusFlowlineVAA_Reachcode_idx ON PlusFlowlineVAA (ReachCode)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS PlusFlowlineVAA_FromMeas_idx ON PlusFlowlineVAA (FromMeas)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS PlusFlowlineVAA_ToMeas_idx ON PlusFlowlineVAA (ToMeas)""")
    
    # Create PlusFlow table and indices
    cursor.execute("""CREATE TABLE IF NOT EXISTS PlusFlow
    (FROMCOMID INTEGER,
    FROMHYDSEQ INTEGER,
    FROMLVLPAT INTEGER,
    TOCOMID INTEGER,
    TOHYDSEQ INTEGER,
    TOLVLPAT INTEGER,
    NODENUMBER INTEGER,
    DELTALEVEL INTEGER,
    DIRECTION INTEGER,
    GAPDISTKM REAL,
    HasGeo TEXT,
    TotDASqKM REAL,
    DivDASqKM REAL)
    """)
    cursor.execute("""CREATE INDEX IF NOT EXISTS plusflow_from_idx ON PlusFlow (FROMCOMID)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS plusflow_to_idx ON PlusFlow (TOCOMID)""")
    
    # Create Gage_Loc table and indices
    cursor.execute("""CREATE TABLE IF NOT EXISTS Gage_Loc
    (ComID INTEGER,
    EventDate DATETIME, 
    ReachCode TEXT,
    ReachSMDat INTEGER,
    Reachresol TEXT,
    FeatureCom INTEGER,
    FeatureCla INTEGER,
    Source_Ori TEXT,
    Source_Dat TEXT,
    Source_Fea TEXT,
    Featuredet TEXT,
    Measure REAL,
    Offset INTEGER,
    EventType TEXT)
    """)
    cursor.execute("""CREATE INDEX IF NOT EXISTS gage_loc_source_fea_idx ON Gage_Loc (Source_Fea)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS reachcode_measure_idx on Gage_Loc (ReachCode,Measure)""")
    
    # Create Gage_Info table
    # Gage_Info.GageID maps to Gage_Loc.Source_Fea
    cursor.execute("""CREATE TABLE IF NOT EXISTS Gage_Info
    (GageID TEXT,
    Agency_cd TEXT,
    Station_NM TEXT,
    State_CD TEXT,
    State TEXT,
    SiteStatus TEXT,
    DA_SQ_Mile REAL,
    Lon_Site REAL,
    Lat_Site REAL,
    Lon_NHD REAL,
    Lat_NHD REAL,
    Reviewed TEXT)
    """)
    cursor.execute("""CREATE INDEX IF NOT EXISTS gage_info_gageID_idx ON Gage_Info (GageID)""")
    
    # Create Gage_Smooth table
    # Gage_Smooth.SITE_NO maps to Gage_Info.GageID
    cursor.execute("""CREATE TABLE IF NOT EXISTS Gage_Smooth
    (SITE_NO TEXT,
    YEAR INTEGER,
    MO INTEGER,
    AVE REAL,
    COMPLETERE REAL)
    """)
    cursor.execute("""CREATE UNIQUE INDEX IF NOT EXISTS gage_smooth_idx ON Gage_Smooth (SITE_NO, YEAR, MO)""")
    
    # Create NHDReachCode_Comid table and indices
    cursor.execute("""CREATE TABLE IF NOT EXISTS NHDReachCode_Comid
    (COMID INTEGER,
    REACHCODE TEXT,
    REACHSMDAT DATETIME,
    RESOLUTION TEXT,
    GNIS_ID INTEGER,
    GNIS_NAME TEXT)
    """)
    cursor.execute("""CREATE INDEX IF NOT EXISTS NHDReachCode_Comid_Comid_idx ON NHDReachCode_Comid (COMID)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS NHDReachCode_Comid_Reachcode_idx ON NHDReachCode_Comid (REACHCODE)""")
    
    # Create NHDFlowline table and indices
    cursor.execute("""CREATE TABLE IF NOT EXISTS NHDFlowline
    (COMID INTEGER,
    FDATE DATETIME,
    RESOLUTION TEXT,
    GNIS_ID INTEGER,
    GNIS_NAME TEXT,
    LENGTHKM REAL,
    REACHCODE TEXT,
    FLOWDIR TEXT,
    WBAREACOMI INTEGER,
    FTYPE TEXT,
    FCODE INTEGER,
    SHAPE_LENG REAL,
    ENABLED TEXT,
    GNIS_NBR INTEGER)
    """)
    cursor.execute("""CREATE INDEX IF NOT EXISTS NHDFlowline_Comid_idx ON NHDReachCode_Comid (COMID)""")
    cursor.execute("""CREATE INDEX IF NOT EXISTS NHDFlowline_Reachcode_idx ON NHDReachCode_Comid (REACHCODE)""")
    
    cursor.close()
    
    # 5. Import NHDPlus data into SQLite database
    # Find PlusFlow.dbf files, open each, import into DB
    print("Importing regional PlusFlowlineVAA.dbf records into CONUS database (this will take a while) ...")
    cursor = conn.cursor()
    dbfs = subprocess.check_output("%s %s -type f -iname PlusFlowlineVAA.dbf -print" % (pathOfFind, args.outputDir,), shell=True).split()
    numFiles = len(dbfs)
    currFile = 0
    for file in dbfs:
        #print file
        pctComplete = (float(currFile) / float(numFiles)) * 100
        currFile = currFile + 1
        sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)" % (currFile, numFiles, pctComplete))
        sys.stdout.flush()
        f = open(file, 'rb')
        db = list(dbfreader(f))
        f.close()
        records = db[2:]
        for record in records:
            #print record
            cursor.execute("""INSERT INTO PlusFlowlineVAA
    (ComID,Fdate,StreamLeve,StreamOrde,StreamCalc,FromNode,ToNode,Hydroseq,LevelPathI,Pathlength,TerminalPa,ArbolateSu,Divergence,StartFlag,TerminalFl,DnLevel,ThinnerCod,UpLevelPat,UpHydroseq,DnLevelPat,DnMinorHyd,DnDrainCou,DnHydroseq,FromMeas,ToMeas,ReachCode,LengthKM,Fcode,RtnDiv,OutDiv,DivEffect,VPUIn,VPUOut,TravTime,PathTime,AreaSqKM,TotDASqKM,DivDASqKM)
    VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)""",
            (record[0], record[1].strftime("%Y-%m-%d %H:%M:%S"), record[2], record[3], record[4], record[5], record[6], record[7], record[8], float(record[9]), record[10], float(record[11]), record[12], record[13], record[14], record[15], record[16], record[17], record[18], record[19], record[20], record[21], record[22], float(record[23]), float(record[24]), unicode(record[25], errors='replace'), float(record[26]), record[27], record[28], record[29], record[30], record[31], record[32], float(record[33]), float(record[34]), float(record[35]), float(record[36]), float(record[37])))
    conn.commit()
    cursor.close()
    
    pctComplete = (float(currFile) / float(numFiles)) * 100
    sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)\n" % (currFile, numFiles, pctComplete)) 
    
    # Find PlusFlow.dbf files, open each, import into DB
    print("Importing regional PlusFlow.dbf records into CONUS database (this will take a while) ...")
    cursor = conn.cursor()
    dbfs = subprocess.check_output("%s %s -type f -iname PlusFlow.dbf -print" % (pathOfFind, args.outputDir,), shell=True).split()
    numFiles = len(dbfs)
    currFile = 0
    for file in dbfs:
        #print file
        pctComplete = (float(currFile) / float(numFiles)) * 100
        currFile = currFile + 1
        sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)" % (currFile, numFiles, pctComplete))
        sys.stdout.flush()
        f = open(file, 'rb')
        db = list(dbfreader(f))
        f.close()
        records = db[2:]
        for record in records:
            #print record
            cursor.execute("""INSERT INTO PlusFlow
    (FROMCOMID,FROMHYDSEQ,FROMLVLPAT,TOCOMID,TOHYDSEQ,TOLVLPAT,NODENUMBER,DELTALEVEL,DIRECTION,GAPDISTKM,HasGeo,TotDASqKM,DivDASqKM)
    VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?)""",
            (record[0], record[1], record[2], record[3], record[4], record[5], record[6], record[7], record[8], float(record[9]), record[10], float(record[11]), float(record[12])))
    conn.commit()
    cursor.close()
    
    pctComplete = (float(currFile) / float(numFiles)) * 100
    sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)\n" % (currFile, numFiles, pctComplete)) 
    
    # Find NHDReachCode_Comid.dbf files, open each, import into DB    
    print("Importing regional NHDReachCode_Comid.dbf records into CONUS database (this will take a while) ...")
    cursor = conn.cursor()
    dbfs = subprocess.check_output("%s %s -type f -iname NHDReachCode_Comid.dbf -print" % (pathOfFind, args.outputDir,), shell=True).split()
    numFiles = len(dbfs)
    currFile = 0
    for file in dbfs:
        #print file
        pctComplete = (float(currFile) / float(numFiles)) * 100
        currFile = currFile + 1
        sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)" % (currFile, numFiles, pctComplete))
        sys.stdout.flush()
        f = open(file, 'rb')
        db = list(dbfreader(f))
        f.close()
        records = db[2:]
        for record in records:
            #print record
            cursor.execute("""INSERT INTO NHDReachCode_Comid
    (COMID,REACHCODE,REACHSMDAT,RESOLUTION,GNIS_ID,GNIS_NAME)
    VALUES (?,?,?,?,?,?)""",
            (record[0], unicode(record[1], errors='replace'), record[2].strftime("%Y-%m-%d %H:%M:%S"), unicode(record[3], errors='replace'), record[4], unicode(record[5], errors='replace')))
    conn.commit()
    cursor.close()
    
    pctComplete = (float(currFile) / float(numFiles)) * 100
    sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)\n" % (currFile, numFiles, pctComplete)) 
    
    # Find NHDFlowline.dbf files, open each, import into DB 
    print("Importing regional NHDFlowline.dbf records into CONUS database (this will take a while) ...")
    cursor = conn.cursor()
    dbfs = subprocess.check_output("%s %s -type f -iname NHDFlowline.dbf -print" % (pathOfFind, args.outputDir,), shell=True).split()
  
    numFiles = len(dbfs)
    currFile = 0
    for file in dbfs:
        #print file
        pctComplete = (float(currFile) / float(numFiles)) * 100
        currFile = currFile + 1
        sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)" % (currFile, numFiles, pctComplete))
        sys.stdout.flush()
        f = open(file, 'rb')
        db = list(dbfreader(f))
        f.close()
        records = db[2:]
        for record in records:
            #print record
            # Handle case where NHDFlowline record lacks GNIS_NBR attribute
            try:
                GNIS_NBR = record[13]
            except IndexError:
                GNIS_NBR = 0
            cursor.execute("""INSERT INTO NHDFlowline
    (COMID,FDATE,RESOLUTION,GNIS_ID,GNIS_NAME,LENGTHKM,REACHCODE,FLOWDIR,WBAREACOMI,FTYPE,FCODE,SHAPE_LENG,ENABLED,GNIS_NBR)
    VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)""",
            (record[0], record[1].strftime("%Y-%m-%d %H:%M:%S"), unicode(record[2], errors='replace'), record[3], unicode(record[4], errors='replace'), float(record[5]), unicode(record[6], errors='replace'), unicode(record[7], errors='replace'), record[8], unicode(record[9], errors='replace'), record[10], float(record[11]), unicode(record[12], errors='replace'), GNIS_NBR))
    conn.commit()
    cursor.close()
    
    pctComplete = (float(currFile) / float(numFiles)) * 100
    sys.stdout.write("\r\tProcessing file %d of %d (%.0f%%)\n" % (currFile, numFiles, pctComplete)) 
    
    # Find GageLoc.dbf file, import into DB 
    print("Importing national GageLoc.dbf ...")
    cursor = conn.cursor()
    dbf = subprocess.check_output("%s %s -type f -iname GageLoc.dbf -print" % (pathOfFind, args.outputDir,), shell=True).split()[0]
    assert(dbf)
    print dbf
    f = open(dbf, 'rb')
    db = list(dbfreader(f))
    f.close()
    records = db[2:]
    for record in records:
        #print record
        cursor.execute("""INSERT INTO Gage_Loc
    (ComID,EventDate,ReachCode,ReachSMDat,Reachresol,FeatureCom,FeatureCla,Source_Ori,Source_Dat,Source_Fea,Featuredet,Measure,Offset,EventType)
    VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)""",
        (record[0], record[1].strftime("%Y-%m-%d %H:%M:%S"), unicode(record[2], errors='replace'), record[3], unicode(record[4], errors='replace'), record[5], record[6], unicode(record[7], errors='replace'), unicode(record[8], errors='replace'), unicode(record[9], errors='replace'), unicode(record[10], errors='replace'), float(record[11]), record[12], unicode(record[13], errors='replace')))
    conn.commit()
    cursor.close()
    
    # Find GageInfo.dbf file, import into DB 
    print("Importing national GageInfo.dbf ...")
    cursor = conn.cursor()
    dbf = subprocess.check_output("%s %s -type f -iname GageInfo.dbf -print" % (pathOfFind, args.outputDir,), shell=True).split()[0]
    assert(dbf)
    print dbf
    f = open(dbf, 'rb')
    db = list(dbfreader(f))
    f.close()
    records = db[2:]
    for record in records:
        #print record
        # Handle presence of undocumented NHD2DAGE_D field (if present)
        if len(record) > 12:
            cursor.execute("""INSERT INTO Gage_Info
        (GageID,Agency_cd,Station_NM,State_CD,State,SiteStatus,DA_SQ_Mile,Lon_Site,Lat_Site,Lon_NHD,Lat_NHD,Reviewed)
        VALUES (?,?,?,?,?,?,?,?,?,?,?,?)""",
            (unicode(record[0],errors='replace'), unicode(record[1], errors='replace'), unicode(record[2], errors='replace'), unicode(record[3], errors='replace'), unicode(record[4], errors='replace'), unicode(record[5], errors='replace'), float(record[6]), float(record[7]), float(record[8]), float(record[9]), float(record[10]), unicode(record[12], errors='replace') ))
        else:
            cursor.execute("""INSERT INTO Gage_Info
        (GageID,Agency_cd,Station_NM,State_CD,State,SiteStatus,DA_SQ_Mile,Lon_Site,Lat_Site,Lon_NHD,Lat_NHD,Reviewed)
        VALUES (?,?,?,?,?,?,?,?,?,?,?,?)""",
            (unicode(record[0],errors='replace'), unicode(record[1], errors='replace'), unicode(record[2], errors='replace'), unicode(record[3], errors='replace'), unicode(record[4], errors='replace'), unicode(record[5], errors='replace'), float(record[6]), float(record[7]), float(record[8]), float(record[9]), float(record[10]), unicode(record[11], errors='replace') ))
    conn.commit()
    cursor.close()
    
    # Find Gage_Smooth.DBF file, import into DB 
    print("Importing national Gage_Smooth.DBF ...")
    cursor = conn.cursor()
    dbf = subprocess.check_output("%s %s -type f -iname Gage_Smooth.DBF -print" % (pathOfFind, args.outputDir,), shell=True).split()[0]
    assert(dbf)
    print dbf
    f = open(dbf, 'rb')
    db = list(dbfreader(f))
    f.close()
    records = db[2:]
    for record in records:
        #print record
        cursor.execute("""INSERT INTO Gage_Smooth
    (SITE_NO,YEAR,MO,AVE,COMPLETERE)
    VALUES (?,?,?,?,?)""",
        (unicode(record[0], errors='replace'), record[1], record[2], float(record[3]), float(record[4]) ))
    conn.commit()
    cursor.close()
    
    conn.close()


#!/usr/bin/env python
"""@package GHCNDSetup
    
@brief Builds SQLite3/Spatialite database needed for querying NCDC Global Historical Climatology Network 
station metadata downloaded from http://www1.ncdc.noaa.gov/pub/data/ghcn/daily/ghcnd-stations.txt.
This database is indexed to allow fast spatial queries of station information. 

@note Requires pyspatialite 3.0.1.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

Usage:
@code
GHCNDSetup.py -o <output_dir>
@endcode
"""
import os, sys, errno
import argparse
import urllib
import re
from pyspatialite import dbapi2 as spatialite


_sanitizeString = '[\'";]'

SRS = int(4326)
DB_NAME = 'GHCND.spatialite'
URL = 'http://www1.ncdc.noaa.gov/pub/data/ghcn/daily/ghcnd-stations.txt'

parser = argparse.ArgumentParser(description='Build database of GHCN station metadata')
parser.add_argument('-o', '--output', dest='outputDir', required=True,
                    help='Directory to which database named "GHCND.sqlite" should be placed')
parser.add_argument('-u', '--url', dest='url', required=False,
                    help='Override station metadata URL')
args = parser.parse_args()

if not os.access(args.outputDir, os.W_OK):
    raise IOError(errno.EACCES, "Not allowed to write to output directory %s" %
                  args.outputDir)

ghcnDB = os.path.join(args.outputDir, DB_NAME)

url = URL
if args.url:
    url = args.url
    
# Delete DB if it already exists
if os.path.exists(ghcnDB):
    os.unlink(ghcnDB)
    
# 1. Create database
conn = spatialite.connect(ghcnDB)
cursor = conn.cursor()

# Setup spatial metadata
sys.stdout.write("Initializing spatial database...")
cursor.execute("""SELECT InitSpatialMetaData()""")
cursor.execute("""SELECT CheckSpatialMetaData()""")
hasSpatial = cursor.fetchone()[0] == 1
if not hasSpatial:
    sys.exit("Failed to create spatial metadata in table %s." % (ghcnDB,) )

# Create station table and indices
cursor.execute("""CREATE TABLE IF NOT EXISTS ghcn_station
(id TEXT NOT NULL PRIMARY KEY,
name TEXT,
elevation_m REAL)
""")
cursor.execute("""SELECT AddGeometryColumn('ghcn_station', 'coord', 4326, 'POINT', 'XY', 1)""")
sys.stdout.write("done\n")
sys.stdout.flush()

# 2. Fetch station metadata
sys.stdout.write("Downloading station data from NCDC (this may take a while)...")
f = urllib.urlopen(url)

# 3. Insert station metadata into database
line = f.readline()
while line:
    line.strip()
    #sys.stdout.write(line)
    id = unicode(line[:11].strip(), errors='replace')
    id = re.sub(_sanitizeString, '', id) # A poor substitute for proper escapeing, but SQLite3 python module sucks in this regard
    lat = float(line[12:19].strip())
    lon = float(line[21:29].strip())
    elev = float(line[31:36].strip())
    name = unicode(line[41:70].strip(), errors='replace')
    name = re.sub(_sanitizeString, '', name) # A poor substitute for proper escapeing, but SQLite3 python module sucks in this regard
    
    # SQLite can't handle parameter substitution within quotes (e.g. GeomFromText('POINT(:lon :lat)')
    #    cursor.execute("INSERT INTO ghcn_station (id,name,elevation_m,coord) VALUES (:id,:name,:elevation_m,GeomFromText('POINT(:lon :lat)', :srs) )",
    #                   {"id": id, "name": name, "elevation_m": elev, "lon": lon, "lat": lat, "srs": SRS})
    # So we have to do it the dangerous way ...
    sql = u"INSERT INTO ghcn_station (id,name,elevation_m,coord) VALUES ('%s','%s',%f,GeomFromText('POINT(%f %f)', %d) )" % \
        (id, name, elev, lon, lat, SRS)
    #print("sql to exec: %s\n" % (sql,) )
    cursor.execute(sql)
    
    line = f.readline()
    
conn.commit()
# Index the data
cursor.execute("""SELECT CreateSpatialIndex('ghcn_station', 'coord')""")
conn.commit()
cursor.close()
conn.close()
sys.stdout.write("done\n")


#!/usr/bin/env python
"""@package GetCatchmentShapefileForHYDRO1kBasins

@brief Extract tile for HYDRO1k digital elevation model (DEM) stored locally

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GDAL/OGR', 'PATH_OF_OGR2OGR'
   'HYDRO1k', 'PATH_OF_HYDRO1K_BAS'
   'HYDRO1k', 'HYDRO1k_BAS_LAYER_NAME'
   
Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   study_area_shapefile [the name of the catchment shapefile] 

Usage:
@code
GetCatchmentShapefileForHYDRO1kBasins.py -p /path/to/project_dir -b <basin1> <basin2> ... <basinN>
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os
import sys
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.hydro1k.basins import getCatchmentShapefileForHYDRO1kBasins

# Handle command line options
parser = argparse.ArgumentParser(description='Get shapefile for the drainage area of an NHDPlus2 streamflow gage')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the catchment shapefile to be written.  File extension ".shp" will be added.  If a file of this name exists, program will silently exit.')
parser.add_argument('-b', '--basins', dest='basins', nargs='+', required=True,
                    help='HYDRO1k basins to extract')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GDAL/OGR', 'PATH_OF_OGR2OGR'):
    sys.exit("Config file %s does not define option %s in section %s" % \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_OGR2OGR'))
if not context.config.has_option('HYDRO1k', 'PATH_OF_HYDRO1K_BAS'):
    sys.exit("Config file %s does not define option %s in section %s" % \
          (args.configfile, 'HYDRO1k', 'PATH_OF_HYDRO1K_BAS'))
if not context.config.has_option('HYDRO1k', 'HYDRO1k_BAS_LAYER_NAME'):
    sys.exit("Config file %s does not define option %s in section %s" % \
          (args.configfile, 'HYDRO1k', 'HYDRO1k_BAS_LAYER_NAME'))
  
if args.outfile:
    outfile = args.outfile
else:
    outfile = "catchment"

shapeFilename = "%s.shp" % (outfile)
shapeFilepath = os.path.join(context.projectDir, shapeFilename)
if not os.path.exists(shapeFilepath):
    getCatchmentShapefileForHYDRO1kBasins(context.config, context.projectDir, shapeFilename, args.basins)
    
    # Write provenance
    asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
    asset.name = 'study_area_shapefile'
    asset.dcIdentifier = shapeFilename
    asset.dcSource = 'http://eros.usgs.gov/#/Find_Data/Products_and_Data_Available/gtopo30/hydro/namerica'
    asset.dcTitle = 'Study area shapefile derived from HYDRO1k Basins'
    asset.dcPublisher = 'USGS'
    asset.dcDescription = cmdline
    asset.writeToMetadata(context)

    # Write processing history
    GenericMetadata.appendProcessingHistoryItem(context, cmdline)
    